# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from featureform.proto import serving_pb2 as featureform_dot_proto_dot_serving__pb2


class FeatureStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.TrainingData = channel.unary_stream(
                '/featureform.serving.proto.Feature/TrainingData',
                request_serializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRow.FromString,
                )
        self.TrainingDataColumns = channel.unary_unary(
                '/featureform.serving.proto.Feature/TrainingDataColumns',
                request_serializer=featureform_dot_proto_dot_serving__pb2.TrainingDataColumnsRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingColumns.FromString,
                )
        self.FeatureServe = channel.unary_unary(
                '/featureform.serving.proto.Feature/FeatureServe',
                request_serializer=featureform_dot_proto_dot_serving__pb2.FeatureServeRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.FeatureRow.FromString,
                )
        self.SourceData = channel.unary_stream(
                '/featureform.serving.proto.Feature/SourceData',
                request_serializer=featureform_dot_proto_dot_serving__pb2.SourceDataRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.SourceDataRow.FromString,
                )
        self.SourceColumns = channel.unary_unary(
                '/featureform.serving.proto.Feature/SourceColumns',
                request_serializer=featureform_dot_proto_dot_serving__pb2.SourceColumnRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.SourceDataColumns.FromString,
                )
        self.Nearest = channel.unary_unary(
                '/featureform.serving.proto.Feature/Nearest',
                request_serializer=featureform_dot_proto_dot_serving__pb2.NearestRequest.SerializeToString,
                response_deserializer=featureform_dot_proto_dot_serving__pb2.NearestResponse.FromString,
                )


class FeatureServicer(object):
    """Missing associated documentation comment in .proto file."""

    def TrainingData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TrainingDataColumns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FeatureServe(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SourceData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SourceColumns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Nearest(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_FeatureServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'TrainingData': grpc.unary_stream_rpc_method_handler(
                    servicer.TrainingData,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.TrainingDataRow.SerializeToString,
            ),
            'TrainingDataColumns': grpc.unary_unary_rpc_method_handler(
                    servicer.TrainingDataColumns,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.TrainingDataColumnsRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.TrainingColumns.SerializeToString,
            ),
            'FeatureServe': grpc.unary_unary_rpc_method_handler(
                    servicer.FeatureServe,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.FeatureServeRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.FeatureRow.SerializeToString,
            ),
            'SourceData': grpc.unary_stream_rpc_method_handler(
                    servicer.SourceData,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.SourceDataRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.SourceDataRow.SerializeToString,
            ),
            'SourceColumns': grpc.unary_unary_rpc_method_handler(
                    servicer.SourceColumns,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.SourceColumnRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.SourceDataColumns.SerializeToString,
            ),
            'Nearest': grpc.unary_unary_rpc_method_handler(
                    servicer.Nearest,
                    request_deserializer=featureform_dot_proto_dot_serving__pb2.NearestRequest.FromString,
                    response_serializer=featureform_dot_proto_dot_serving__pb2.NearestResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'featureform.serving.proto.Feature', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Feature(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def TrainingData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/featureform.serving.proto.Feature/TrainingData',
            featureform_dot_proto_dot_serving__pb2.TrainingDataRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.TrainingDataRow.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TrainingDataColumns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/featureform.serving.proto.Feature/TrainingDataColumns',
            featureform_dot_proto_dot_serving__pb2.TrainingDataColumnsRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.TrainingColumns.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def FeatureServe(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/featureform.serving.proto.Feature/FeatureServe',
            featureform_dot_proto_dot_serving__pb2.FeatureServeRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.FeatureRow.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SourceData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/featureform.serving.proto.Feature/SourceData',
            featureform_dot_proto_dot_serving__pb2.SourceDataRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.SourceDataRow.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SourceColumns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/featureform.serving.proto.Feature/SourceColumns',
            featureform_dot_proto_dot_serving__pb2.SourceColumnRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.SourceDataColumns.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Nearest(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/featureform.serving.proto.Feature/Nearest',
            featureform_dot_proto_dot_serving__pb2.NearestRequest.SerializeToString,
            featureform_dot_proto_dot_serving__pb2.NearestResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



import base64
import hashlib
import os
import threading
from abc import ABC, abstractmethod
from typing import Optional

import featureform as ff
import requests
from flask import Flask, request
from typeguard import typechecked
from werkzeug.serving import make_server


@typechecked
class AuthConfig(ABC):
    @abstractmethod
    def get_authorization_endpoint(self, redirect_uri: str, code_challenge: str) -> str:
        pass

    @abstractmethod
    def get_token_exchange_endpoint(self) -> str:
        pass


@typechecked
class OktaAuthConfig(AuthConfig):
    def __init__(
        self, domain: str, authorization_server_id: str, client_id: str
    ) -> None:
        self.domain = domain
        self.authorization_server_id = authorization_server_id
        self.client_id = client_id

    def get_authorization_endpoint(self, redirect_uri: str, code_challenge: str) -> str:
        return (
            f"https://{self.domain}/oauth2/{self.authorization_server_id}/v1/authorize?client_id={self.client_id}"
            f"&response_type=code&scope=openid&redirect_uri={redirect_uri}&state=random_state"
            f"&code_challenge_method=S256&code_challenge={code_challenge}"
        )

    def get_token_exchange_endpoint(self) -> str:
        return f"https://{self.domain}/oauth2/{self.authorization_server_id}/v1/token"


@typechecked
class AuthService(ABC):
    def __init__(self, auth_config) -> None:
        self._auth_config = auth_config

    @abstractmethod
    def authenticate(self) -> None:
        pass

    @abstractmethod
    def get_access_token(self) -> Optional[str]:
        pass

    @abstractmethod
    def clear_access_token(self) -> None:
        pass


@typechecked
class PassThroughService(AuthService):
    def authenticate(self) -> None:
        pass

    def get_access_token(self) -> Optional[str]:
        pass

    def clear_access_token(self) -> None:
        pass


@typechecked
class OktaOAuth2PKCE(AuthService):
    def __init__(self, auth_config: OktaAuthConfig) -> None:
        super().__init__(auth_config)
        # up casts the instance in this class to OktaAuthConfig
        self._auth_config: OktaAuthConfig = auth_config
        self.redirect_uri = "http://localhost:9080/authorization-code/callback"
        self._access_token = None
        self._code_verifier = None
        self._callback_server = None
        self._callback_server_thread = None
        self._callback_flask_app = Flask(__name__)
        self._auth_completed = threading.Event()

        @self._callback_flask_app.route("/authorization-code/callback")
        def callback():
            auth_code = request.args.get("code")
            threading.Thread(
                target=self._exchange_code_for_token, args=(auth_code,)
            ).start()
            return "Authentication successful! You can close this window."

    def authenticate(self) -> None:
        self._code_verifier = self._create_code_verifier()
        code_challenge = self._create_code_challenge(self._code_verifier)
        auth_url = self._auth_config.get_authorization_endpoint(
            self.redirect_uri, code_challenge
        )
        print(f"Please visit the following URL to authenticate: {auth_url}")
        self._callback_server_thread = threading.Thread(
            target=self._start_callback_server
        )
        self._callback_server_thread.start()
        self._auth_completed.wait()

    def get_access_token(self) -> Optional[str]:
        return self._access_token

    def clear_access_token(self) -> None:
        self._auth_completed.clear()
        self._access_token = None

    @staticmethod
    def _create_code_verifier():
        token = base64.urlsafe_b64encode(os.urandom(40)).decode("utf-8")
        return token.rstrip("=")

    @staticmethod
    def _create_code_challenge(verifier):
        m = hashlib.sha256()
        m.update(verifier.encode("utf-8"))
        challenge = base64.urlsafe_b64encode(m.digest()).decode("utf-8")
        return challenge.rstrip("=")

    def _exchange_code_for_token(self, auth_code):
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        data = {
            "grant_type": "authorization_code",
            "code": auth_code,
            "redirect_uri": self.redirect_uri,
            "client_id": self._auth_config.client_id,
            "code_verifier": self._code_verifier,
        }
        try:
            response = requests.post(
                self._auth_config.get_token_exchange_endpoint(),
                headers=headers,
                data=data,
            )
            if response.status_code == 200:
                print("Authentication Succeeded!")
                self._access_token = response.json().get("access_token")
                self._auth_completed.set()
            else:
                raise Exception("Authentication Failed.")
        finally:
            self._stop_callback_server()

    def _start_callback_server(self):
        self._callback_server = make_server("127.0.0.1", 9080, self._callback_flask_app)
        self._callback_server.serve_forever()

    def _stop_callback_server(self):
        if self._callback_server:
            self._callback_server.shutdown()


@typechecked
class OktaOAuth2ClientCredentials(AuthService):
    def __init__(self, auth_config) -> None:
        super().__init__(auth_config)
        self._access_token = None

    def authenticate(self) -> None:
        client_id = os.environ.get("FF_OAUTH_CLIENT_ID")
        client_secret = os.environ.get("FF_OAUTH_CLIENT_SECRET")
        if not client_id or not client_secret:
            print("No client credentials found in Environment")
            return None
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }
        data = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
        }
        response = requests.post(
            self._auth_config.get_token_exchange_endpoint(), headers=headers, data=data
        )
        self._access_token = (
            response.json().get("access_token") if response.status_code == 200 else None
        )
        if self._access_token:
            print("Authentication Succeeded!")
        else:
            print("Failed to authenticate with client credentials")

    def get_access_token(self) -> Optional[str]:
        return self._access_token

    def clear_access_token(self) -> None:
        self._access_token = None


@typechecked
class AuthenticationManagerImpl:
    TOKEN_FILENAME = "access_token.txt"

    def __init__(self) -> None:
        self._access_token = None
        self._auth_config = None
        self._services = []
        feature_form_dir = os.environ.get("FEATUREFORM_DIR", ".featureform")
        self.auth_dir = os.path.join(feature_form_dir, "auth")
        os.makedirs(self.auth_dir, exist_ok=True)
        self.token_filepath = os.path.join(self.auth_dir, self.TOKEN_FILENAME)

    def _write_token_to_file(self, token: str):
        with open(self.token_filepath, "w") as f:
            f.write(token)
        os.chmod(self.token_filepath, 0o600)

    def _read_token_from_file(self) -> Optional[str]:
        if os.path.exists(self.token_filepath):
            with open(self.token_filepath, "r") as f:
                return f.read().strip()
        return None

    def delete_expired_token(self):
        self._access_token = None
        for service in self._services:
            service.clear_access_token()
        feature_form_dir = os.environ.get("FEATUREFORM_DIR", ".featureform")
        auth_dir = os.path.join(feature_form_dir, "auth")
        token_filepath = os.path.join(
            auth_dir, AuthenticationManagerImpl.TOKEN_FILENAME
        )
        if os.path.exists(token_filepath):
            os.remove(token_filepath)

    def get_access_token_or_authenticate(self, insecure, host) -> Optional[str]:
        if not self._access_token:
            self._access_token = self._read_token_from_file()
            if not self._access_token:
                if not self._services:
                    self._auth_config = self._load_auth_config(insecure, host)
                    if self._auth_config is not None:
                        self._services = [
                            OktaOAuth2ClientCredentials(self._auth_config),
                            OktaOAuth2PKCE(self._auth_config),
                        ]
                    else:
                        self._services = [PassThroughService(self._auth_config)]

                for service in self._services:
                    service.authenticate()
                    token = service.get_access_token()
                    if token:
                        self._access_token = token
                        self._write_token_to_file(token)
                        break
        return self._access_token

    @staticmethod
    def _load_auth_config(insecure, host) -> Optional[OktaAuthConfig]:
        config = ff.Client(insecure=insecure, host=host).get_auth_config()
        if config.WhichOneof("config") == "okta":
            okta_config = config.okta
            return OktaAuthConfig(
                domain=okta_config.domain,
                authorization_server_id=okta_config.authorization_server_id,
                client_id=okta_config.client_id,
            )
        elif config.WhichOneof("config") == "pass_through":
            return None


singleton = AuthenticationManagerImpl()



from unittest.mock import patch, Mock

import pytest

from featureform.lib import auth, OktaAuthConfig, OktaOAuth2PKCE, OktaOAuth2ClientCredentials


def test_okta_auth_config():
    config = OktaAuthConfig("domain.com", "server", "client123")
    expected_endpoint = "https://domain.com/oauth2/server/v1/authorize?client_id=client123&response_type=code&scope=openid&redirect_uri=some_uri&state=random_state&code_challenge_method=S256&code_challenge=some_challenge"
    assert config.get_authorization_endpoint("some_uri", "some_challenge") == expected_endpoint
    assert config.get_token_exchange_endpoint() == "https://domain.com/oauth2/server/v1/token"


def test_code_verifier_creation():
    config = OktaAuthConfig("domain.com", "server", "client123")
    service = OktaOAuth2PKCE(config)
    verifier = service._create_code_verifier()
    assert len(verifier) == 54  # This assumes a 40-byte token as input


def test_code_challenge_creation():
    config = OktaAuthConfig("domain.com", "server", "client123")
    service = OktaOAuth2PKCE(config)
    verifier = "some_verifier"
    challenge = service._create_code_challenge(verifier)
    assert len(challenge) <= 44  # Base64 url-safe encoded SHA256


@patch("requests.post")
def test_exchange_code_for_token_success(mock_post):
    config = OktaAuthConfig("domain.com", "server", "client123")
    service = OktaOAuth2PKCE(config)

    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "access_token": "some_access_token"
    }
    mock_post.return_value = mock_response

    service._exchange_code_for_token("some_auth_code")
    assert service.get_access_token() == "some_access_token"


@patch("requests.post")
def test_exchange_code_for_token_fail(mock_post):
    config = OktaAuthConfig("domain.com", "server", "client123")
    service = OktaOAuth2PKCE(config)

    mock_response = Mock()
    mock_response.status_code = 400
    mock_response.json.return_value = {}
    mock_post.return_value = mock_response

    with pytest.raises(Exception) as e:
        service._exchange_code_for_token("some_auth_code")
    assert str(e.value) == "Authentication Failed."
    assert service.get_access_token() is None


@patch("requests.post")
@patch("os.environ")
def test_client_credentials_success(mock_environ, mock_post):
    mock_environ.get.side_effect = ["client_id", "client_secret"]
    config = OktaAuthConfig("domain.com", "server", "client123")
    service = OktaOAuth2ClientCredentials(config)

    mock_response = Mock()
    mock_response.status_code = 200
    mock_response.json.return_value = {
        "access_token": "some_access_token"
    }
    mock_post.return_value = mock_response

    service.authenticate()
    assert service.get_access_token() == "some_access_token"


@patch("requests.post")
@patch("os.environ")
def test_client_credentials_fail(mock_environ, mock_post):
    mock_environ.get.side_effect = [None, None]
    config = OktaAuthConfig("domain.com", "server", "client123")
    service = OktaOAuth2ClientCredentials(config)

    mock_response = Mock()
    mock_response.status_code = 400
    mock_response.json.return_value = {}
    mock_post.return_value = mock_response

    service.authenticate()
    assert service.get_access_token() is None


@patch.object(OktaOAuth2PKCE, 'authenticate')
@patch.object(OktaOAuth2PKCE, 'get_access_token', return_value="token")
@patch.object(auth.singleton, '_load_auth_config', return_value=OktaAuthConfig(
    domain="",
    authorization_server_id="",
    client_id="",
))
def test_authentication_manager_success(mock_get_access_token, mock_authenticate, mocked_load):
    assert auth.singleton._access_token is None
    token = auth.singleton.get_access_token_or_authenticate(insecure=True)
    assert token == "token"
    auth.singleton._access_token = "override_with_cache_token"
    token_cached = auth.singleton.get_access_token_or_authenticate(insecure=True)
    assert token_cached == "override_with_cache_token"


# Demo program

def execute_dummy_method():
    print("Executing dummy method...")


if __name__ == "__main__":
    access_token = None

    while True:
        print("\nChoose an action:")
        print("1. Execute dummy method")
        print("2. Exit")
        choice = input("> ")

        if choice == '1':
            if not access_token:
                access_token = auth.singleton.get_access_token_or_authenticate(insecure=True)
                if not access_token:
                    print("Failed to authenticate.")
                    continue
            execute_dummy_method()

        elif choice == '2':
            break



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

import click
from .client import Client
from .list import *
from .get import *
import os
from flask import Flask
from .dashboard_metadata import dashboard_app
import validators
import urllib.request
from .version import get_package_version
from .tls import get_version_local, get_version_hosted

resource_types = [
    "feature",
    "source",
    "training-set",
    "label",
    "entity",
    "provider",
    "model",
    "user",
]

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    # fmt: off
    """
    \b
    ______         _                   __                     
    |  ___|       | |                 / _|                    
    | |_ ___  __ _| |_ _   _ _ __ ___| |_ ___  _ __ _ __ ___  
    |  _/ _ \/ _` | __| | | | '__/ _ \  _/ _ \| '__| '_ ` _ \ 
    | ||  __/ (_| | |_| |_| | | |  __/ || (_) | |  | | | | | |
    \_| \___|\__,_|\__|\__,_|_|  \___|_| \___/|_|  |_| |_| |_|

    Interact with Featureform's Feature Store via the official command line interface.
    """
    # fmt: on
    pass


@cli.command()
@click.option(
    "--host",
    "host",
    required=False,
    help="The host address of the API server to connect to",
)
@click.option(
    "--cert", "cert", required=False, help="Path to self-signed TLS certificate"
)
@click.option("--insecure", is_flag=True, help="Disables TLS verification")
@click.option("--local", is_flag=True, help="Enables local mode")
@click.argument("resource_type", required=True)
@click.argument("name", required=True)
@click.argument("variant", required=False)
def get(host, cert, insecure, local, resource_type, name, variant):
    """Get resources of a given type."""
    if local:
        if host != None:
            raise ValueError("Cannot be local and have a host")

    elif host == None:
        host = os.getenv("FEATUREFORM_HOST")
        if host == None:
            raise ValueError(
                "Host value must be set with --host flag or in env as FEATUREFORM_HOST"
            )

    client = Client(host=host, local=local, insecure=insecure, cert_path=cert)

    resource_get_functions_variant = {
        "feature": client.print_feature,
        "label": client.print_label,
        "source": client.print_source,
        "trainingset": client.print_training_set,
        "training-set": client.print_training_set,
    }

    resource_get_functions = {
        "user": client.get_user,
        "model": client.get_model,
        "entity": client.get_entity,
        "provider": client.get_provider,
    }

    if resource_type in resource_get_functions_variant:
        resource_get_functions_variant[resource_type](
            name=name, variant=variant, local=local
        )
    elif resource_type in resource_get_functions:
        resource_get_functions[resource_type](name=name, local=local)
    else:
        raise ValueError("Resource type not found")


@cli.command()
@click.option(
    "--host",
    "host",
    required=False,
    help="The host address of the API server to connect to",
)
@click.option(
    "--cert", "cert", required=False, help="Path to self-signed TLS certificate"
)
@click.option("--insecure", is_flag=True, help="Disables TLS verification")
@click.option("--local", is_flag=True, help="Enable local mode")
@click.argument("resource_type", required=True)
def list(host, cert, insecure, local, resource_type):
    if local:
        if host != None:
            raise ValueError("Cannot be local and have a host")

    elif host == None:
        host = os.getenv("FEATUREFORM_HOST")
        if host == None:
            raise ValueError(
                "Host value must be set with --host flag or in env as FEATUREFORM_HOST"
            )

    client = Client(host=host, local=local, insecure=insecure, cert_path=cert)

    resource_list_functions = {
        "features": client.list_features,
        "labels": client.list_labels,
        "sources": client.list_sources,
        "trainingsets": client.list_training_sets,
        "training-sets": client.list_training_sets,
        "users": client.list_users,
        "models": client.list_models,
        "entities": client.list_entities,
        "providers": client.list_providers,
    }

    if resource_type in resource_list_functions:
        resource_list_functions[resource_type](local=local)
    else:
        raise ValueError("Resource type not found")


app = Flask(__name__)
app.register_blueprint(dashboard_app)


@cli.command()
@click.option("--local", is_flag=True, help="Required for local mode only")
def version(local):
    client_version = get_package_version()
    host = os.getenv("FEATUREFORM_HOST", "")
    cluster_version = ""
    output = f"Client Version: {client_version}"
    if local == False:
        try:
            cluster_version = get_version_hosted(host)
        except:
            cluster_version = "Cannot retrieve: Check your FEATUREFORM_HOST value. If using local mode, use the --local flag."
        output += f"\nCluster Version: {cluster_version}"

    print(output)


@cli.command()
def dash():
    run_dashboard()


@cli.command()
def dashboard():
    run_dashboard()


def run_dashboard():
    app.run(threaded=True, port=os.getenv("LOCALMODE_DASHBOARD_PORT", 3000))


@cli.command()
@click.argument("files", required=True, nargs=-1)
@click.option(
    "--host",
    "host",
    required=False,
    help="The host address of the API server to connect to",
)
@click.option(
    "--cert", "cert", required=False, help="Path to self-signed TLS certificate"
)
@click.option("--insecure", is_flag=True, help="Disables TLS verification")
@click.option("--local", is_flag=True, help="Enable local mode")
@click.option(
    "--dry-run", is_flag=True, help="Checks the definitions without applying them"
)
@click.option("--no-wait", is_flag=True, help="Applies the resources asynchronously")
@click.option("--debug", is_flag=True, help="Enable verbose errors")
@click.option(
    "--verbose", is_flag=True, help="Prints all errors at the end of an apply"
)
def apply(host, cert, insecure, local, files, dry_run, no_wait, debug, verbose):
    for file in files:
        if os.path.isfile(file):
            read_file(file)
        elif validators.url(file):
            read_url(file)
        else:
            raise ValueError(
                f"Argument must be a path to a file or URL with a valid schema (http:// or https://): {file}"
            )

    client = Client(
        host=host,
        local=local,
        insecure=insecure,
        cert_path=cert,
        dry_run=dry_run,
        debug=debug,
    )
    asynchronous = no_wait
    client.apply(asynchronous=asynchronous, verbose=verbose)


@cli.command()
@click.option(
    "--query",
    "-q",
    "query",
    required=True,
    help="The phrase to search resources (e.g. 'quick').",
)
@click.option(
    "--host",
    "host",
    required=False,
    help="The host address of the API server to connect to",
)
@click.option(
    "--cert", "cert", required=False, help="Path to self-signed TLS certificate"
)
@click.option("--insecure", is_flag=True, help="Disables TLS verification")
@click.option("--local", is_flag=True, help="Enable local mode")
@click.option("--debug", is_flag=True, help="Enable verbose errors")
def search(query, host, cert, insecure, local, debug):
    client = Client(
        host=host, local=local, insecure=insecure, cert_path=cert, debug=debug
    )
    results = client.search(query, local)
    if local:
        format_rows("NAME", "VARIANT", "TYPE")
        for r in results:
            desc = (
                r["description"][:cutoff_length] + "..."
                if len(r["description"]) > 0
                else ""
            )
            format_rows(r["name"], r["variant"], r["resource_type"])


def read_file(file):
    with open(file, "r") as py:
        exec_file(py, file)


def read_url(url):
    try:
        with urllib.request.urlopen(url) as py:
            exec_file(py, url)
    except Exception as e:
        raise ValueError(f"Could not apply the provided URL: {e}: {url}")


def exec_file(file, name):
    code = compile(file.read(), name, "exec")
    # Create a new global namespace for each file to ensure that
    # global variables, such as `ff`, are not undefined in the
    # context of class attribute assignments (e.g. `label = ff.Label()`)
    file_globals = {}
    exec(code, file_globals)


if __name__ == "__main__":
    cli()



import sys

import grpc

from .enums import ComputationMode
from .lib import auth


class BCOLORS:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


class StubExceptionWrapper:
    def __init__(self, wrapped, debug=False, insecure=False, host=None):
        self._insecure = insecure
        self._host = host
        self.wrapped = wrapped
        self.debug = debug
        # Store metadata as an instance variable

    @staticmethod
    def _merge_auth_metadata(original_metadata, auth_metadata):
        if original_metadata is None:
            return auth_metadata
        # Remove existing authorization header if present
        merged_metadata = [
            item for item in original_metadata if item[0].lower() != "authorization"
        ]
        # Append the new authorization header
        merged_metadata.extend(auth_metadata)
        return merged_metadata

    @staticmethod
    def is_streaming_response(obj):
        return hasattr(obj, "__iter__") and not isinstance(
            obj, (str, bytes, dict, list)
        )

    def streaming_wrapper(self, multi_threaded_rendezvous):
        try:
            for message in multi_threaded_rendezvous:
                yield message
        except grpc.RpcError as e:
            # Handle the error gracefully here.
            self.handle_grpc_error(e)

    def __getattr__(self, name):
        insecure = self._insecure
        attr = getattr(self.wrapped, name)
        if name != "GetAuthConfig" and callable(attr):

            def wrapper(*args, **kwargs):
                try:
                    # Use the stored metadata for the call
                    token = auth.singleton.get_access_token_or_authenticate(
                        self._insecure, self._host
                    )
                    if token is not None:
                        kwargs["metadata"] = self._merge_auth_metadata(
                            kwargs.get("metadata"),
                            [("authorization", "Bearer " + token)],
                        )
                    result = attr(*args, **kwargs)
                    # If result is a streaming call, wrap it.
                    if self.is_streaming_response(result):
                        return self.streaming_wrapper(result)
                    return result
                except grpc.RpcError as e:
                    self.handle_grpc_error(e)

            return wrapper
        else:
            return attr

    def handle_grpc_error(self, e):
        sys.tracebacklimit = 0
        initial_ex = None
        if self.debug:
            initial_ex = e
            sys.tracebacklimit = None
        if e.code() == grpc.StatusCode.UNAUTHENTICATED:
            auth.singleton.delete_expired_token()
            raise Exception(
                f"{BCOLORS.FAIL}Authentication failed.{BCOLORS.ENDC}\n"
                "Your access token is no longer valid. Please re-run the previous command to authenticate."
            )

        elif e.code() == grpc.StatusCode.UNAVAILABLE:
            print("\n")
            raise Exception(
                f"{BCOLORS.FAIL}Could not connect to Featureform.{BCOLORS.ENDC}\n"
                "Please check if your FEATUREFORM_HOST and FEATUREFORM_CERT environment variables are set "
                "correctly or are explicitly set in the client or command line.\n"
                f"Details: {e.details()}"
            ) from initial_ex
        elif e.code() == grpc.StatusCode.UNKNOWN:
            raise Exception(
                f"{BCOLORS.FAIL}Error: {e.details()}{BCOLORS.ENDC}"
            ) from initial_ex
        else:
            raise e


class InvalidTrainingSetFeatureComputationMode(Exception):
    def __init__(
        self,
        feature_name,
        feature_variant,
        mode=ComputationMode.CLIENT_COMPUTED.value,
        message=None,
    ):
        if message is None:
            message = (
                f"Feature '{feature_name}:{feature_variant}' is on demand. "
                f"Cannot use {mode} features for training sets. "
            )

        Exception.__init__(self, message)


class FeatureNotFound(Exception):
    def __init__(self, feature_name, feature_variant, message=None):
        error_message = f"Feature '{feature_name}:{feature_variant}' not found. Verify that the feature is registered."

        if message is not None:
            error_message = f"{error_message} {message}"

        Exception.__init__(self, error_message)


class LabelNotFound(Exception):
    def __init__(self, label_name, label_variant, message=None):
        error_message = f"Label '{label_name}:{label_variant}' not found. Verify that the label is registered."
        if message is not None:
            error_message = f"{error_message} {message}"

        Exception.__init__(self, error_message)


class InvalidSQLQuery(Exception):
    def __init__(self, query, message=None):
        error_message = f"Invalid SQL query. Query: ' {query} '"
        if message is not None:
            error_message = f"{error_message} {message}"

        Exception.__init__(self, error_message)
